@@ -30,8 +30,7 @@
 #  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #
 
-__version__ = '3.1.2'
-from .ast import DataType
+__version__ = '3.3.0'
 
 from .engine import resolve_attribute
 from .engine import resolve_item
@@ -44,3 +43,5 @@ from .errors import EngineError
 from .errors import EvaluationError
 from .errors import RuleSyntaxError
 from .errors import SymbolResolutionError
+
+from .types import DataType
