@@ -32,6 +32,7 @@
 
 import collections
 import collections.abc
+import contextlib
 import datetime
 import decimal
 import functools
@@ -42,6 +43,7 @@ import threading
 from . import ast
 from . import errors
 from . import parser
+from .suggestions import suggest_symbol
 
 import dateutil.tz
 
@@ -53,7 +55,7 @@ def _today(builtins):
 
 def _tls_getter(thread_local, key, _builtins):
 	# a function stub to be used with functools.partial for retrieving thread-local values
-	return thread_local.storage.get(key)
+	return getattr(thread_local.storage, key)
 
 def resolve_attribute(thing, name):
 	"""
@@ -70,7 +72,7 @@ def resolve_attribute(thing, name):
 	:return: The value for the corresponding attribute *name*.
 	"""
 	if not hasattr(thing, name):
-		raise errors.SymbolResolutionError(name, thing=thing)
+		raise errors.SymbolResolutionError(name, thing=thing, suggestion=suggest_symbol(name, dir(thing)))
 	return getattr(thing, name)
 
 def resolve_item(thing, name):
@@ -83,15 +85,15 @@ def resolve_item(thing, name):
 	:param str name: The symbol name that is being resolved.
 	:return: The value for the corresponding attribute *name*.
 	"""
-	if not isinstance(thing, collections.abc.Iterable):
+	if not isinstance(thing, collections.abc.Mapping):
 		raise errors.SymbolResolutionError(name, thing=thing)
 	if name not in thing:
-		raise errors.SymbolResolutionError(name, thing=thing)
+		raise errors.SymbolResolutionError(name, thing=thing, suggestion=suggest_symbol(name, thing.keys()))
 	return thing[name]
 
 def _type_resolver(type_map, name):
 	if name not in type_map:
-		raise errors.SymbolResolutionError(name)
+		raise errors.SymbolResolutionError(name, suggestion=suggest_symbol(name, type_map.keys()))
 	return type_map[name]
 
 def _float_op(value, op):
@@ -153,7 +155,7 @@ class _AttributeResolver(object):
 			raise errors.AttributeResolutionError(name, object_type, thing=thing)
 		resolver = attribute_resolvers.get(name)
 		if resolver is None:
-			raise errors.AttributeResolutionError(name, object_type, thing=thing)
+			raise errors.AttributeResolutionError(name, object_type, thing=thing, suggestion=suggest_symbol(name, attribute_resolvers.keys()))
 		return resolver
 
 	def resolve_type(self, object_type, name):
@@ -166,6 +168,10 @@ class _AttributeResolver(object):
 		"""
 		return self._get_resolver(object_type, name).result_type
 
+	@attribute('to_epoch', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
+	def datetime_to_epoch(self, value):
+		return value.timestamp()
+
 	@attribute('date', ast.DataType.DATETIME, result_type=ast.DataType.DATETIME)
 	def datetime_date(self, value):
 		return value.replace(hour=0, minute=0, second=0, microsecond=0)
@@ -260,7 +266,7 @@ class _AttributeResolver(object):
 		value = value.strip()
 		if re.match(r'-?inf', value):
 			return decimal.Decimal(value)
-		match = re.match(r'^(' + parser.Parser.t_FLOAT + ')$', value)
+		match = re.match(r'^(' + parser.Parser.get_token_regex('FLOAT') + ')$', value)
 		if match is None:
 			return decimal.Decimal('nan')
 		return parser.literal_eval(match.group(0))
@@ -359,6 +365,21 @@ class Builtins(collections.abc.Mapping):
 		default_value_types.update(kwargs.pop('value_types', {}))
 		return cls(default_values, value_types=default_value_types, **kwargs)
 
+class _ThreadLocalStorage(object):
+	"""
+	An object whose attributes are required to be tracked separately among multiple threads. This is to guarantee that
+	if a context is used by one or more rules that are being evaluated simultaneously in multiple threads, that the
+	states are kept isolated.
+	"""
+	__slots__ = ('assignment_scopes', 'regex_groups')
+	def __init__(self):
+		self.assignment_scopes = collections.deque()
+		self.regex_groups = None
+
+	def reset(self):
+		self.assignment_scopes.clear()
+		self.regex_groups = None
+
 class Context(object):
 	"""
 	An object defining the context for a rule's evaluation. This can be used to change the behavior of certain aspects
@@ -424,7 +445,7 @@ class Context(object):
 		self.default_value = default_value
 		"""The *default_value* parameter from :py:meth:`~__init__`"""
 		self.builtins = Builtins.from_defaults(
-			values={'re_groups': functools.partial(_tls_getter, self._thread_local, 'regex.groups')},
+			values={'re_groups': functools.partial(_tls_getter, self._thread_local, 'regex_groups')},
 			value_types={'re_groups': ast.DataType.ARRAY(ast.DataType.STRING)},
 			timezone=default_timezone
 		)
@@ -436,10 +457,25 @@ class Context(object):
 		self.__type_resolver = type_resolver or (lambda _: ast.DataType.UNDEFINED)
 		self.__resolver = resolver or resolve_item
 
+	@contextlib.contextmanager
+	def assignments(self, *assignments):
+		"""
+		Add the specified assignments to a thread-specific scope. This is used when an assignment originates from an
+		expression.
+
+		:param assignment: The one or more assignments to define.
+		:type assignment: :py:class:`~rule_engine.ast.Assignment`
+		"""
+		self._tls.assignment_scopes.append({assign.name: assign for assign in assignments})
+		try:
+			yield
+		finally:
+			self._tls.assignment_scopes.pop()
+
 	@property
 	def _tls(self):
 		if not hasattr(self._thread_local, 'storage'):
-			self._thread_local.storage = {}
+			self._thread_local.storage = _ThreadLocalStorage()
 		return self._thread_local.storage
 
 	def resolve(self, thing, name, scope=None):
@@ -461,6 +497,9 @@ class Context(object):
 		if isinstance(thing, Builtins):
 			return resolve_item(thing, name)
 		if scope is None:
+			for assignments in self._tls.assignment_scopes:
+				if name in assignments:
+					return assignments[name].value
 			return self.__resolver(thing, name)
 		raise errors.SymbolResolutionError(name, symbol_scope=scope, thing=thing)
 
@@ -494,6 +533,9 @@ class Context(object):
 		"""
 		if scope == Builtins.scope_name:
 			return self.builtins.resolve_type(name)
+		for assignments in self._tls.assignment_scopes:
+			if name in assignments:
+				return assignments[name].value_type
 		return self.__type_resolver(name)
 
 class Rule(object):
@@ -562,7 +604,7 @@ class Rule(object):
 		:return: The value the rule evaluates to. Unlike the :py:meth:`.matches` method, this is not necessarily a
 			boolean.
 		"""
-		self.context._tls.clear()
+		self.context._tls.reset()
 		with decimal.localcontext(self.context.decimal_context):
 			return self.statement.evaluate(thing)
 
