@@ -30,9 +30,10 @@
 #  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #
 
-import ast as ast_
+import ast as pyast
 import collections
 import threading
+import types as pytypes
 
 from . import ast
 from . import errors
@@ -40,7 +41,21 @@ from . import errors
 import ply.lex as lex
 import ply.yacc as yacc
 
-literal_eval = ast_.literal_eval
+literal_eval = pyast.literal_eval
+
+class _DeferredAstNode(object):
+	__slots__ = ('cls', 'args', 'kwargs', 'method')
+	def __init__(self, cls, *, args, kwargs=None, method='build'):
+		if not issubclass(cls, ast.ASTNodeBase):
+			raise TypeError('cls is not a subclass of AstNodeBase')
+		self.cls = cls
+		self.args = args
+		self.kwargs = kwargs or {}
+		self.method = method
+
+	def build(self):
+		constructor = getattr(self.cls, self.method)
+		return constructor(*self.args, **self.kwargs)
 
 class ParserBase(object):
 	"""
@@ -69,8 +84,9 @@ class ParserBase(object):
 
 	def parse(self, text, context, **kwargs):
 		"""
-		Parse the specified text in an abstract syntax tree of nodes that can
-		later be evaluated.
+		Parse the specified text in an abstract syntax tree of nodes that can later be evaluated. This is done in two
+		phases. First, the syntax is parsed and a tree of deferred / uninitialized AST nodes are constructed. Next each
+		node is built recursively using it's respective :py:meth:`rule_engine.ast.ASTNodeBase.build`.
 
 		:param str text: The grammar text to parse into an AST.
 		:param context: A context for specifying parsing and evaluation options.
@@ -81,9 +97,11 @@ class ParserBase(object):
 		kwargs['lexer'] = kwargs.pop('lexer', self._lexer)
 		with self.__mutex:
 			self.context = context
+			# phase 1: parse the string into a tree of deferred nodes
 			result = self._parser.parse(text, **kwargs)
 			self.context = None
-		return result
+		# phase 2: initialize each AST node recursively, providing them with an opportunity to define assignments
+		return result.build()
 
 class Parser(ParserBase):
 	"""
@@ -111,6 +129,7 @@ class Parser(ParserBase):
 		'<':   'LT',    '<=': 'LE',
 		# logical operators
 		'and': 'AND',   'or': 'OR',     'not': 'NOT',
+		'for': 'FOR',   'if': 'IF',
 		# other operators
 		'.':   'ATTR',
 		'&.':  'ATTR_SAFE',
@@ -130,6 +149,8 @@ class Parser(ParserBase):
 		'in': 'IN',
 		'or': 'OR',
 		'not': 'NOT',
+		'for': 'FOR',
+		'if': 'IF'
 	}
 	tokens = (
 		'DATETIME', 'FLOAT', 'STRING', 'SYMBOL',
@@ -181,6 +202,21 @@ class Parser(ParserBase):
 		('left',     'ATTR', 'ATTR_SAFE'),
 	)
 
+	@classmethod
+	def get_token_regex(cls, token_name):
+		"""
+		Return the regex that is used by the specified token.
+
+		:param str token_name: The token for which to return the regex.
+		:rtype: str
+		"""
+		obj = getattr(cls, 't_' + token_name, None)
+		if isinstance(obj, str):
+			return obj
+		elif isinstance(obj, pytypes.FunctionType):
+			return obj.__doc__
+		raise ValueError('unknown token: ' + token_name)
+
 	def t_POW(self, t):
 		r'\*\*?'
 		if t.value == '*':
@@ -228,7 +264,7 @@ class Parser(ParserBase):
 
 	def t_SYMBOL(self, t):
 		r'\$?[a-zA-Z_][a-zA-Z0-9_]*'
-		if t.value in ('if', 'elif', 'else', 'for', 'while'):
+		if t.value in ('elif', 'else', 'while'):
 			raise errors.RuleSyntaxError("syntax error (the {} keyword is reserved for future use)".format(t.value))
 		t.type = self.reserved_words.get(t.value, 'SYMBOL')
 		return t
@@ -246,7 +282,7 @@ class Parser(ParserBase):
 
 	def p_statement_expr(self, p):
 		'statement : expression'
-		p[0] = ast.Statement(self.context, p[1])
+		p[0] = _DeferredAstNode(ast.Statement, args=(self.context, p[1]))
 
 	def p_expression_getattr(self, p):
 		"""
@@ -254,7 +290,7 @@ class Parser(ParserBase):
 		       | object ATTR_SAFE SYMBOL
 		"""
 		op_name = self.op_names.get(p[2])
-		p[0] = ast.GetAttributeExpression(self.context, p[1], p[3], safe=op_name == 'ATTR_SAFE').reduce()
+		p[0] = _DeferredAstNode(ast.GetAttributeExpression, args=(self.context, p[1], p[3]), kwargs={'safe': op_name == 'ATTR_SAFE'})
 
 	def p_expression_object(self, p):
 		"""
@@ -267,7 +303,7 @@ class Parser(ParserBase):
 		expression : expression QMARK expression COLON expression
 		"""
 		condition, _, case_true, _, case_false = p[1:6]
-		p[0] = ast.TernaryExpression(self.context, condition, case_true, case_false).reduce()
+		p[0] = _DeferredAstNode(ast.TernaryExpression, args=(self.context, condition, case_true, case_false))
 
 	def p_expression_arithmetic(self, p):
 		"""
@@ -281,7 +317,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.ArithmeticExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.ArithmeticExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_bitwise(self, p):
 		"""
@@ -291,7 +327,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.BitwiseExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.BitwiseExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_bitwise_shift(self, p):
 		"""
@@ -300,7 +336,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.BitwiseShiftExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.BitwiseShiftExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_contains(self, p):
 		"""
@@ -309,11 +345,11 @@ class Parser(ParserBase):
 		"""
 		if len(p) == 4:
 			member, _, container = p[1:4]
-			p[0] = ast.ContainsExpression(self.context, container, member).reduce()
+			p[0] = _DeferredAstNode(ast.ContainsExpression, args=(self.context, container, member))
 		else:
 			member, _, _, container = p[1:5]
-			p[0] = ast.ContainsExpression(self.context, container, member).reduce()
-			p[0] = ast.UnaryExpression(self.context, 'NOT', p[0]).reduce()
+			p[0] = _DeferredAstNode(ast.ContainsExpression, args=(self.context, container, member))
+			p[0] = _DeferredAstNode(ast.UnaryExpression, args=(self.context, 'NOT', p[0]))
 
 	def p_expression_comparison(self, p):
 		"""
@@ -322,7 +358,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.ComparisonExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.ComparisonExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_arithmetic_comparison(self, p):
 		"""
@@ -333,7 +369,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.ArithmeticComparisonExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.ArithmeticComparisonExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_fuzzy_comparison(self, p):
 		"""
@@ -344,7 +380,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.FuzzyComparisonExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.FuzzyComparisonExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_logic(self, p):
 		"""
@@ -353,7 +389,7 @@ class Parser(ParserBase):
 		"""
 		left, op, right = p[1:4]
 		op_name = self.op_names[op]
-		p[0] = ast.LogicExpression(self.context, op_name, left, right).reduce()
+		p[0] = _DeferredAstNode(ast.LogicExpression, args=(self.context, op_name, left, right))
 
 	def p_expression_group(self, p):
 		'object : LPAREN expression RPAREN'
@@ -361,7 +397,7 @@ class Parser(ParserBase):
 
 	def p_expression_negate(self, p):
 		'expression : NOT expression'
-		p[0] = ast.UnaryExpression(self.context, 'NOT', p[2]).reduce()
+		p[0] = _DeferredAstNode(ast.UnaryExpression, args=(self.context, 'NOT', p[2]))
 
 	def p_expression_symbol(self, p):
 		'object : SYMBOL'
@@ -370,12 +406,12 @@ class Parser(ParserBase):
 		if name[0] == '$':
 			scope = 'built-in'
 			name = name[1:]
-		p[0] = ast.SymbolExpression(self.context, name, scope=scope).reduce()
+		p[0] = _DeferredAstNode(ast.SymbolExpression, args=(self.context, name), kwargs={'scope': scope})
 
 	def p_expression_uminus(self, p):
 		'expression : SUB expression %prec UMINUS'
 		names = {'-': 'UMINUS'}
-		p[0] = ast.UnaryExpression(self.context, names[p[1]], p[2]).reduce()
+		p[0] = _DeferredAstNode(ast.UnaryExpression, args=(self.context, names[p[1]], p[2]))
 
 	# Literal expressions
 	def p_expression_boolean(self, p):
@@ -383,39 +419,39 @@ class Parser(ParserBase):
 		expression : TRUE
 				   | FALSE
 		"""
-		p[0] = ast.BooleanExpression(self.context, p[1] == 'true')
+		p[0] = _DeferredAstNode(ast.BooleanExpression, args=(self.context, p[1] == 'true'))
 
 	def p_expression_datetime(self, p):
 		'object : DATETIME'
-		p[0] = ast.DatetimeExpression.from_string(self.context, literal_eval(p[1]))
+		p[0] = _DeferredAstNode(ast.DatetimeExpression, args=(self.context, literal_eval(p[1])), method='from_string')
 
 	def p_expression_float(self, p):
 		'expression : FLOAT'
-		p[0] = ast.FloatExpression(self.context, float(literal_eval(p[1])))
+		p[0] = _DeferredAstNode(ast.FloatExpression, args=(self.context, float(literal_eval(p[1]))))
 
 	def p_expression_float_nan(self, p):
 		'expression : FLOAT_NAN'
-		p[0] = ast.FloatExpression(self.context, float('nan'))
+		p[0] = _DeferredAstNode(ast.FloatExpression, args=(self.context, float('nan')))
 
 	def p_expression_float_inf(self, p):
 		'expression : FLOAT_INF'
-		p[0] = ast.FloatExpression(self.context, float('inf'))
+		p[0] = _DeferredAstNode(ast.FloatExpression, args=(self.context, float('inf')))
 
 	def p_expression_null(self, p):
 		'object : NULL'
 		# null is an object because of the safe operator
-		p[0] = ast.NullExpression(self.context)
+		p[0] = _DeferredAstNode(ast.NullExpression, args=(self.context,))
 
 	def p_expression_set(self, p):
 		"""
 		object : LBRACE ary_members RBRACE
 			   | LBRACE ary_members COMMA RBRACE
 		"""
-		p[0] = ast.SetExpression(self.context, tuple(p[2])).reduce()
+		p[0] = _DeferredAstNode(ast.SetExpression, args=(self.context, tuple(p[2])))
 
 	def p_expression_string(self, p):
 		'object : STRING'
-		p[0] = ast.StringExpression(self.context, literal_eval(p[1]))
+		p[0] = _DeferredAstNode(ast.StringExpression, args=(self.context, literal_eval(p[1])))
 
 	def p_expression_array(self, p):
 		"""
@@ -424,9 +460,19 @@ class Parser(ParserBase):
 			   | LBRACKET ary_members COMMA RBRACKET
 		"""
 		if len(p) < 4:
-			p[0] = ast.ArrayExpression(self.context, tuple()).reduce()
+			p[0] = _DeferredAstNode(ast.ArrayExpression, args=(self.context, tuple()))
 		else:
-			p[0] = ast.ArrayExpression(self.context, tuple(p[2])).reduce()
+			p[0] = _DeferredAstNode(ast.ArrayExpression, args=(self.context, tuple(p[2])))
+
+	def p_expression_array_comprehension(self, p):
+		"""
+		object : LBRACKET expression FOR SYMBOL IN expression RBRACKET
+			   | LBRACKET expression FOR SYMBOL IN expression IF expression RBRACKET
+		"""
+		condition = None
+		if len(p) == 10:
+			condition = p[8]
+		p[0] = _DeferredAstNode(ast.ComprehensionExpression, args=(self.context, p[2], p[4], p[6]), kwargs={'condition': condition})
 
 	def p_expression_array_members(self, p):
 		"""
@@ -448,9 +494,9 @@ class Parser(ParserBase):
 			   | LBRACE map_members COMMA RBRACE
 		"""
 		if len(p) < 4:
-			p[0] = ast.MappingExpression(self.context, tuple()).reduce()
+			p[0] = _DeferredAstNode(ast.MappingExpression, args=(self.context, tuple()))
 		else:
-			p[0] = ast.MappingExpression(self.context, tuple(p[2])).reduce()
+			p[0] = _DeferredAstNode(ast.MappingExpression, args=(self.context, tuple(p[2])))
 
 	def p_expression_mapping_member(self, p):
 		"""
@@ -470,7 +516,7 @@ class Parser(ParserBase):
 		object : object LBRACKET expression RBRACKET
 		"""
 		container, lbracket, item = p[1:4]
-		p[0] = ast.GetItemExpression(self.context, container, item, safe=lbracket == '&[').reduce()
+		p[0] = _DeferredAstNode(ast.GetItemExpression, args=(self.context, container, item), kwargs={'safe': lbracket == '&['})
 
 	def p_expression_getslice(self, p):
 		"""
@@ -492,4 +538,4 @@ class Parser(ParserBase):
 			start, _, stop = p[3:6]
 		else:
 			raise errors.RuleSyntaxError('invalid get slice expression')
-		p[0] = ast.GetSliceExpression(self.context, container, start, stop, safe=safe).reduce()
+		p[0] = _DeferredAstNode(ast.GetSliceExpression, args=(self.context, container, start, stop), kwargs={'safe': safe})
